/* This is a set of functions that provides minimal filesystem
 * functionality to the Linux bootstrapper.  All we can do is
 * open and read files... but that's all we need 8-)
 */

#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/string.h>
#include "miniboot.h"
#include "boot_e2lib.h"

#ifndef CONFIG_MINI_LINUX_IN_MEMORY

#define MAX_SUPPORTED_GROUPS	128	/* Supports 1Gb filesystem */
#define MAX_OPEN_FILES		4 

int fd = -1;
struct ext2_super_block sb;
struct ext2_group_desc gds[MAX_SUPPORTED_GROUPS];
int ngroups = 0;
int blocksize;				/* Block size of this fs */
int directlim;				/* Maximum direct blkno */
int ind1lim;				/* Maximum single-indir blkno */
int ind2lim;				/* Maximum double-indir blkno */
int ptrs_per_blk;			/* ptrs/indirect block */
char filename[256];
char blkbuf[EXT2_MAX_BLOCK_SIZE];
int cached_iblkno = -1;
char iblkbuf[EXT2_MAX_BLOCK_SIZE];
int cached_diblkno = -1;
char diblkbuf[EXT2_MAX_BLOCK_SIZE];

struct inode_table_entry {
    struct ext2_inode inode;
    int inumber;
    int free;
    unsigned short old_mode;
} inode_table[MAX_OPEN_FILES];

/* forward declarations */
void ext2_bread(int fd, long blkno, char *buffer);
void ext2_breadi(struct ext2_inode *ip, long blkno, char *buffer);

/* Initialize an ext2 filesystem; this is sort-of the same idea as
 * "mounting" it.  Read in the relevant control structures and 
 * make them available to the user.  Returns 0 if successful, -1 on
 * failure.
 */

#if 0
#define DEBUG_EXT2 1
#endif
int ext2_mount(long device)
{
    int i;
    int sb_offset;
    int sb_block = 1;

#ifdef DEBUG_EXT2
    printk("ext2_mount(): called, device = 0x%04x\n", device);
#endif
    /* Initialize the inode table */
    for (i = 0; i < MAX_OPEN_FILES; i++) {
	inode_table[i].free = 1;
	inode_table[i].inumber = 0;
    }

    /* Read in the first superblock */
    sb_offset = sb_block * EXT2_MIN_BLOCK_SIZE;

#ifdef DEBUG_EXT2
    printk("Reading the first superblock() at offset 0x%p\n", sb_offset);
#endif
    if (device_read(fd, (char *) &sb, sizeof(sb), sb_offset) !=
      sizeof(sb)) {
	printk("ext2 sb read failed");
	device_close(fd);
	return (-1);
    }

    if (sb.s_magic != EXT2_SUPER_MAGIC) {
	printk("ext2_mount: bad magic 0x%x\n", sb.s_magic);
	device_close(fd);
	return (-1);
    }

#ifdef DEBUG_EXT2
    printk("Superblock\n");
    printk("   [%d] inodes\n", sb.s_inodes_count);
    printk("   [%d] blocks\n", sb.s_blocks_count);
    printk("   [%d] blocks per group\n", sb.s_blocks_per_group);
#endif

    ngroups =
      (sb.s_blocks_count + sb.s_blocks_per_group - 1) / sb.s_blocks_per_group;
#ifdef DEBUG_EXT2
    printk("...ngroups is %d\n", ngroups);
#endif
    if (ngroups > MAX_SUPPORTED_GROUPS) {
	printk("ext2_mount: too many groups (%d): max is %d\n", 
	  ngroups, MAX_SUPPORTED_GROUPS);
	device_close(fd);
	return (-1);
    }

    /* Read in the group descriptors (immediately follows superblock) */
#ifdef DEBUG_EXT2
    printk("...reading in the group descriptors (%d) at offset 0x%p\n", ngroups,
	   sb_offset + sizeof(sb));
    printk("   size is %d\n", ngroups * sizeof(struct ext2_group_desc));
#endif
#if 1
    if (device_read(fd, (char *) gds, ngroups * sizeof(struct ext2_group_desc), 
		sb_offset + sizeof(sb)) != ngroups * sizeof(struct ext2_group_desc)) {
	printk("ext2_mount: could not read groups\n");
	device_close(fd);
	return(-1);
    }
#else
    if (device_read(fd, (char *) gds, 1024, 
		sb_offset + sizeof(sb)) != 1024) {
	printk("ext2_mount: could not read groups\n");
	device_close(fd);
	return(-1);
    }
#endif



#ifdef DEBUG_EXT2
    for (i = 0; i < ngroups; i++) {
	printk("Group %d\n", i);
	printk("   [0x%X] blocks bitmap\n", gds[i].bg_block_bitmap);
	printk("   [0x%X] inode bitmap\n", gds[i].bg_inode_bitmap);
	printk("   [0x%X] inode table\n", gds[i].bg_inode_table);
    }
#endif

    /* Calculate direct/indirect block limits for this file system (blocksize
     * dependent) */
    blocksize = EXT2_BLOCK_SIZE(&sb);
    directlim = EXT2_NDIR_BLOCKS - 1;
    ptrs_per_blk = blocksize / sizeof(unsigned int);
    ind1lim = ptrs_per_blk + directlim;
    ind2lim = (ptrs_per_blk * ptrs_per_blk) + directlim;

#ifdef DEBUG_EXT2
    printk("returning successfully\n");
#endif
    return (0);
}

long ext2_fs_size(void)
{
    return (sb.s_blocks_count * blocksize);
}


/* Call this when we're all done with the file system.  Since we're not
 * writing to the file system, this will do nothing (no updates needed)
 */
void ext2_finish(void)
{
    device_close(fd);
}



/* Read the specified inode from the disk and return it to the user.
 * Returns NULL if the inode can't be read...
 */
static struct ext2_inode *ext2_iget(int ino)
{
    int i;
    struct ext2_inode *ip;
    struct inode_table_entry *itp;
    int group;
    int offset;

#ifdef DEBUG_EXT2
    printk("ext2_iget(): called for ino = %d\n", ino);
    printk("...inodes per group = %d\n", sb.s_inodes_per_group);
#endif

    ip = NULL;
    itp = NULL;
    for (i = 0; i < MAX_OPEN_FILES; i++) {
	if (inode_table[i].free) {
	    itp = &(inode_table[i]);
	    ip = &(itp->inode);
	    break;
	}
    }
    if ((ip == NULL) || (itp == NULL)) {
	printk("ext2_iget: no free inodes\n");
	return (NULL);
    }

    group = ino / sb.s_inodes_per_group;
    offset = (gds[group].bg_inode_table * blocksize)
      + (((ino - 1) % sb.s_inodes_per_group) * sizeof(struct ext2_inode));

#ifdef DEBUG_EXT2
    printk("...group = %d, gds[group].bg_inode_table = %d\n", group,
	   gds[group].bg_inode_table);
    printk("...reading inode at offset 0x%p\n", offset);
#endif
    if (device_read(fd, (char *) ip, sizeof(struct ext2_inode), offset)
      != sizeof(struct ext2_inode)) {
	printk("ext2_iget: read error\n");
	return (NULL);
    }

    itp->free = 0;
    itp->inumber = ino;
    itp->old_mode = ip->i_mode;

#ifdef DEBUG_EXT2
    printk("ext2_iget(): returning inode, size is %d\n", ip->i_size);
#endif

    return (ip);
}

/* Release our hold on an inode.  Since this is a read-only application,
 * don't worry about putting back any changes...
 */
static void ext2_iput(struct ext2_inode *ip)
{
    struct inode_table_entry *itp;

    /* Find and free the inode table slot we used... */
    itp = (struct inode_table_entry *) ip;

    itp->inumber = 0;
    itp->free = 1;
}


/* Map a block offset into a file into an absolute block number.
 * (traverse the indirect blocks if necessary).  Note: Double-indirect
 * blocks allow us to map over 64Mb on a 1k file system.  Therefore, for
 * our purposes, we will NOT bother with triple indirect blocks.
 *
 * The "allocate" argument is set if we want to *allocate* a block
 * and we don't already have one allocated.
 */
static int ext2_blkno(struct ext2_inode *ip, int blkoff, int allocate)
{
    unsigned int *lp;
    unsigned int *ilp;
    unsigned int *dlp;
    int blkno;
    int iblkno;
    int diblkno;


    ilp = (unsigned int *) iblkbuf;
    dlp = (unsigned int *) diblkbuf;

    if (allocate) {
	printk("ext2_blkno: Cannot allocate on a readonly file system!\n");
	return (0);
    }

    lp = (unsigned int *) blkbuf;

    /* If it's a direct block, it's easy! */
    if (blkoff <= directlim) {
	return (ip->i_block[blkoff]);
    }

    /* Is it a single-indirect? */
    if (blkoff <= ind1lim) {
	iblkno = ip->i_block[EXT2_IND_BLOCK];

	if (iblkno == 0) {
	    return (0);
	}

	/* Read the indirect block */
	if (cached_iblkno != iblkno) {
#ifdef DEBUG_EXT2
	    printk("Reading indirect block at 0x%p\n", (iblkno * blocksize));
#endif
	    if (device_read(fd, iblkbuf, blocksize, (iblkno * blocksize)) !=
	      blocksize) {
		printk("ext2_blkno: error on iblk read\n");
		return (0);
	    }
	    cached_iblkno = iblkno;
	}

	blkno = ilp[blkoff - (directlim + 1)];
	return (blkno);
    }

    /* Is it a double-indirect? */
    if (blkoff <= ind2lim) {

	/* Find the double-indirect block */
	diblkno = ip->i_block[EXT2_DIND_BLOCK];

	if (diblkno == 0) {
	    return (0);
	}

	/* Read in the double-indirect block */
	if (cached_diblkno != diblkno) {
#ifdef DEBUG_EXT2
	    printk("Reading double indirect block at 0x%p\n", 
		   (diblkno * blocksize));
#endif
	    if (device_read(fd, diblkbuf, blocksize, (diblkno * blocksize)) !=
	      blocksize) {
		printk("ext2_blkno: err reading dindr blk\n");
		return (0);
	    }
	    cached_diblkno = diblkno;
	}

	/* Find the single-indirect block pointer ... */
	iblkno = dlp[(blkoff - (ind1lim + 1)) / ptrs_per_blk];

	if (iblkno == 0) {
	    return (0);
	}


	/* Read the indirect block */

	if (cached_iblkno != iblkno) {
	    if (device_read(fd, iblkbuf, blocksize, (iblkno * blocksize)) !=
	      blocksize) {
		printk("ext2_blkno: err on iblk read\n");
		return (0);
	    }
	    cached_iblkno = iblkno;
	}

	/* Find the block itself. */
	blkno = ilp[(blkoff - (ind1lim + 1)) % ptrs_per_blk];
	return (blkno);
    }

    if (blkoff > ind2lim) {
	printk("ext2_blkno: block number too large: %d\n", blkoff);
	return (0);
    }

    return (0);
}




/* Read block number "blkno" from the specified file */
void ext2_bread(int fd, long blkno, char *buffer)
{
    struct ext2_inode *ip;

    ip = &(inode_table[fd].inode);
    ext2_breadi(ip, blkno, buffer);
}

void ext2_breadi(struct ext2_inode *ip, long blkno, char *buffer)
{
    long dev_blkno;


#ifdef DEBUG_EXT2
    printk("ext2_breadi(): called for block %d\n", blkno);
#endif

    dev_blkno = ext2_blkno(ip, blkno, 0);
#ifdef DEBUG_EXT2
    printk("ext2_breadi(): dev_blkno = %d\n", dev_blkno);
#endif
    if (dev_blkno == 0) {

	/* This is a "hole" */
	memset(buffer, 0, blocksize);
    }
      else {

	/* Read it for real */
#ifdef DEBUG_EXT2
	printk("ext2_breadi(): reading block %d at offset 0x%p\n", blkno,
		 (dev_blkno * blocksize));
#endif
	if (device_read(fd, buffer, blocksize, (dev_blkno * blocksize)) !=
	  blocksize) {
	    printk("ext2_bread: read error");
	}
    }
}


static struct ext2_inode *ext2_namei(char *name)
{
    char namebuf[256];
    char *component;
    struct ext2_inode *dir_inode;
    struct ext2_dir_entry *dp;
    int next_ino;


#ifdef DEBUG_EXT2
    printk("ext2_namei(): called for %s\n", name);
#endif

    /* Squirrel away a copy of "namebuf" that we can molest */
    strcpy(namebuf, name);

    /* Start at the root... */
    dir_inode = ext2_iget(EXT2_ROOT_INO);
    if (!dir_inode) return NULL;

    component = strtok(namebuf, "/");
    while (component) {
	int diroffset;
	int blockoffset;
	int component_length;

	/* Search for the specified component in the current directory inode.
	 */

	next_ino = -1;

	component_length = strlen(component);
	diroffset = 0;
#ifdef DEBUG_EXT2
	printk("...dir_inode->i_size = %d\n", dir_inode->i_size);
#endif
	while (diroffset < dir_inode->i_size) {
	    blockoffset = 0;
#ifdef DEBUG_EXT2
	    printk("ext2_namei(): calling ext2_breadi\n");
#endif
	    ext2_breadi(dir_inode, diroffset / blocksize, blkbuf);
	    while (blockoffset < blocksize) {
		dp = (struct ext2_dir_entry *) (blkbuf + blockoffset);
#if 0
	        {
		    int i;

		    printk("...name is ");
		    for (i = 0; i < dp->name_len; i++) printk("%c", dp->name[i]);
		    printk("\n");
	        } 
#endif
		if ((dp->name_len == component_length) && 
		  (strncmp(component, dp->name, component_length) == 0)) {

		    /* Found it! */
		    next_ino = dp->inode;
		    break;
		}

		/* Go to next entry in this block */
		blockoffset += dp->rec_len;
	    }
	    if (next_ino >= 0) {
		break;
	    }

	    /* If we got here, then we didn't find the component. Try the next 
	     * block in this directory... */
	    diroffset += blocksize;
	}  /* end-while */

	/* At this point, we're done with this directory whether we've
	 * succeeded or failed... */
	ext2_iput(dir_inode);

	/* If next_ino is negative, then we've failed (gone all the way through
	 * without finding anything) */
	if (next_ino < 0) {
#ifdef DEBUG_EXT2
	    printk("...we failed\n");
#endif	    
	    return (NULL);
	}

	/* Otherwise, we can get this inode and find the next component
	 * string... */
	dir_inode = ext2_iget(next_ino);

	component = strtok(NULL, "/");
    }

    /* If we get here, then we got through all the components. Whatever we got 
     * must match up with the last one. */
    return (dir_inode);
}

int ext2_open(char *filename)
{

    /* Unix-like open routine.  Returns a small integer (actually an index into
     * the inode table... */
    struct ext2_inode *ip;

#ifdef DEBUG_EXT2
    printk("ext2_open() called with filename = %s\n", filename);
#endif
    ip = ext2_namei(filename);
#ifdef DEBUG_EXT2
    printk("...namei returned %d\n", ip);
#endif

    if (ip) {
	struct inode_table_entry *itp;

	itp = (struct inode_table_entry *) ip;
	return (itp - inode_table);
    }
      else {
	return (-1);
    }
}


void ext2_close(int fd)
{
    ext2_iput(&(inode_table[fd].inode));
    device_close(fd);
}

#endif
