/* Routines to access the NVRAM on NoName/EB64/EB66.  The NVRAM on these
 * platforms is a strange creature... it is essentially a bank-switched
 * device, with 8Kb of data divided into 32 256-byte banks.  The bank
 * number is selected by writing it to I/O port address 0xC00.  The byte
 * within that bank is then selected by reading or writing to an I/O port
 * in the range 0x800 - 0x8ff.
 */

#include <linux/config.h>

#ifdef MINI_NVRAM

#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/bios32.h>

#include "miniboot.h"

#define NVRAM_ADDRHI	0xC00
#define NVRAM_ADDRLO	0x800

/* Copy data in from NVRAM.
 *
 *	offset - Offset in the NVRAM to start copying
 *	dest   - Destination address for the data
 *	size   - Number of bytes to copy.
 */
void nvram_copyin(int offset, char * dest, int size)
{
    int		i;
    char	*dp;
    char	ubcs_b;
    char	verify;

    /* We have enable I/O to the NVRAM ports.  To do this,
     * we have to set an enable bit in the SIO's Utility
     * Bus Chip Select Enable register.
     */
    pcibios_read_config_byte(0, 0x38, 0x4f, &ubcs_b);
    ubcs_b |= 0x80;
    pcibios_write_config_byte(0, 0x38, 0x4f, ubcs_b);
    pcibios_read_config_byte(0, 0x38, 0x4f, &verify);
    if(verify != ubcs_b) {
	printk("nvram_copyin: SIO config failed: expect 0x%x got 0x%x\n",
			ubcs_b, verify);
    }

    for(i = 0, dp=dest; i < size; i++, dp++, offset++) {
	outb((offset >> 8) & 0x1f, NVRAM_ADDRHI);
	*dp = inb(NVRAM_ADDRLO + (offset & 0xff));
    }
}

/* Copy data out to NVRAM.
 *
 *	offset - Offset in the NVRAM to start copying
 *	src    - Address of the data to copy out.
 *	size   - Number of bytes to copy.
 */
void nvram_copyout(int offset, char * src, int size)
{
    int		i;
    char	*dp;
    char	ubcs_b;
    char	verify;

    /* We have enable I/O to the NVRAM ports.  To do this,
     * we have to set an enable bit in the SIO's Utility
     * Bus Chip Select Enable register.
     */
    pcibios_read_config_byte(0, 0x38, 0x4f, &ubcs_b);
    ubcs_b |= 0x80;
    pcibios_write_config_byte(0, 0x38, 0x4f, ubcs_b);
    pcibios_read_config_byte(0, 0x38, 0x4f, &verify);
    if(verify != ubcs_b) {
	printk("nvram_copyin: SIO config failed: expect 0x%x got 0x%x\n",
			ubcs_b, verify);
    }


    for(i = 0, dp=src; i < size; i++, dp++, offset++) {
	outb((offset >> 8) & 0x1f, NVRAM_ADDRHI);
	outb(*dp, NVRAM_ADDRLO + (offset & 0xff));
    }
}

#endif
